import sqlalchemy as sa
import sqlalchemy.ext.declarative as declarative
import sqlalchemy.orm as orm
from zope.sqlalchemy import ZopeTransactionExtension

# Global variables initialized by ``reset()``.
_base = _session = _engines = _zte = None

def reset():
    """Delete all engines and restore the initial module state.
    
    This function is mainly for unit tests and debugging. It undoes all
    customizations and reverts to the initial module state.
    """
    global _base, _session, _engines, _zte
    _zte = ZopeTransactionExtension()
    sm = orm.sessionmaker(extension=[_zte])
    _base = declarative.declarative_base()
    _session = orm.scoped_session(sm)
    _engines = {}

reset()

# PUBLIC API

__all__ = [
    "add_engine", 
    "config_session",
    "get_base",
    "get_session", 
    "get_engine",
    ]
    # "reset" is not included because it's not intended for normal use.

def add_engine(settings=None, name="default", prefix="sqlalchemy.",
    engine=None, **engine_args):
    """Configure a SQLAlchemy database engine and return it.

    I configure an engine in different ways depending on the combination of
    arguments. If ``name`` is not specified or its value is "default", I also
    bind the scoped session and the declarative base's metadata to it.

    Arguments:

    * ``settings``: A dict of application settings (e.g., as parsed from an INI
      file), or a dict containing engine args. If this argument is passed, I
      call ``sqlalchemy.engine_from_config(settings, prefix)``.

    * ``name``: The engine name. This is used to retrieve the engine later. The
      default name is "default".

    * ``prefix``: This is used with ``settings`` to calcuate the engine args.
      The default value is "sqlalchemy.", which tells SQLAlchemy to use the
      keys starting with "sqlalchemy." for this engine.

    * ``engine``: An existing SQLAlchemy engine. If you pass this you can't
      pass ``settings``, ``prefix``, or ``**engine_args``.

    * ``**engine_args``: Engine args supplied as keyword arguments. If
      ``settings`` is also passed, the keyword args override their
      corresponding settings. If ``settings`` is not passed, I call
      ``sqlalchemy.create_engine(**engine_args)`` directly.

    SQLAlchemy will raise a ``KeyError`` if the database URL is not specified.
    This may indicate the settings dict has no "PREFIX.url" key or that the
    ``url`` keyword arg was not passed.

    Examples::

        # Configure engine using a settings dict
        settings = {"sqlalchemy.url": "mysql://..."}
        engine = add_engine(settings, prefix="sqlalchemy.")

        # Configure engine via keyword args
        engine = add_engine(url="mysql://...")

        # ``e`` is an existing SQLAlchemy engine
        engine = add_engine(e)

        # Configure two engines, the first one as default
        settings = {"db1.url": "mysql://...", "db2.url": "postgresql://..."})
        engine1 = add_engine(settings, prefix="db1.")
        engine2 = add_engine(settings, name="stats", prefix="db2.")

        # Configure two engines with no default engine
        settings = {"db1.url": "mysql://...", "db2.url": "postgresql://..."})
        engine1 = add_engine(settings, name="engine1", prefix="db1.")
        engine2 = add_engine(settings, name="engine2", prefix="db2.")
    """
    if engine and (settings or engine_args):
        m = "can't specify settings or engine args when ``engine`` is present"
        raise TypeError(m)
    elif engine:
        e = engine
    elif settings:
        if not prefix:
            raise ValueError("empty prefix ('') is not allowed")
        url_key = prefix + "url"
        if url_key not in settings and "url" not in engine_args:
            msg = """\
no database URL specified
settings key '%surl' is required when using prefix='%s'"""
            msg %= (url_key, prefix)
            if prefix and not prefix.endswith("."):
                msg += "\nHint: did you mean prefix='%s.'?" % prefix
            raise ValueError(msg)
        e = sa.engine_from_config(settings, prefix, **engine_args)
    else:
        try:
            url = engine_args.pop("url")
        except KeyError:
            raise TypeError("must pass settings dict or ``url`` keyword arg")
        e = sa.create_engine(url, **engine_args)
    _engines[name] = e
    if name == "default":
        _session.configure(bind=e)
        _base.metadata.bind = e
    return e

def get_session():
    """Return the central SQLAlchemy scoped session."""
    return _session

def get_engine(name="default"):
    """Return a database engine previously configured with ``add_engine``.

    If no argument, return the default engine. If an engine name is passed,
    return the engine that was registered under that name.

    Raise ``RuntimeError`` if no engine by that name was configured.
    """
    try:
        return _engines[name]
    except KeyError:
        raise RuntimeError("No engine '%s' was configured" % name)

def get_base():
    """Return the central declarative base.
    """
    return _base



from paste.util.template import paste_script_template_renderer
from pyramid.paster import PyramidTemplate

class PyramidSQLAProjectTemplate(PyramidTemplate):
    _template_dir = "pyramid_sqla"
    summary = "Pyramid SQLAlchemy project with view handlers"
    template_renderer = staticmethod(paste_script_template_renderer)

    def pre(self, command, output_dir, vars):
        """Called before template is applied."""
        PyramidTemplate.pre(self, command, output_dir, vars)
        package_logger = vars['package']
        if package_logger == 'root':
            # Rename the app logger in the rare case a project is named 'root'
            package_logger = 'app'
        vars['package_logger'] = package_logger



import os

import pkg_resources
from pyramid.view import static

def add_static_route(config, package, subdir, cache_max_age=3600,
    **add_route_args):
    """Add a route and view to serve static files from a directory.

    I create a catchall route that serves all URLs from a directory of static
    files if the corresponding file exists. Subdirectories are also handled.
    For example, the URL "/robots.txt" corresponds to file
    "PACKAGE/SUBDIR/robots.txt", and "/images/header/logo.png"
    corresponds to "PACKAGE/SUBDIR/images/header/logo.png".  If the file
    doesn't exist, the route won't match the URL, and Pyramid will continue to
    the next route or traversal. The route name is 'static', which must not
    conflict with your application's other routes.

    Usage in __init__.py::

        add_static_route(config, "myapp", "static")

    This serves URLs from the "static" directory in package "myapp".

    Arguments:

    * ``config``: a ``pyramid.config.Configurator`` instance.

    * ``package``: the name of the Python package containing the static files.

    * ``subdir``: the subdirectory in the package that contains the files.
      This should be a relative directory with '/' separators regardless of
      platform. 

    * ``directory``: a Pyramid asset specification ("mypackage:static") or an
      absolute directory. An asset specification is preferred because it will
      automatically adjust if the application is moved to another directory.

    * ``cache_max_age``: influences the ``Expires`` and ``Max-Age``
      response headers returned by the view (default is 3600 seconds or five
      minutes).

    * ``**add_route_args``: additional arguments to ``config.add_route``.
      'name' defaults to "static" but can be overridden. (Every route in your
      application must have a unique name.) 'pattern' and 'view' may not be
      specified and will raise TypeError if they are.
    """
    for bad_arg in ["pattern", "view"]:
        if bad_arg in add_route_args:
            raise TypeError("keyword arg '%s' not allowed")
    name = add_route_args.pop("name", "static")
    pattern = "/*subpath"
    asset = "%s:%s" % (package, subdir)
    view = static(asset, cache_max_age)
    pred = StaticViewPredicate(package, subdir)
    custom_predicates = add_route_args.pop("custom_predicates", [])
    custom_predicates = list(custom_predicates)
    custom_predicates.insert(0, pred)
    config.add_route(name, pattern, view, 
        custom_predicates=custom_predicates, **add_route_args)

#### Private stuff

class StaticViewPredicate(object):
    def __init__(self, package, subdir):
        self.package = package
        self.subdir = subdir

    def __call__(self, info, request):
        import logging
        log = logging.getLogger(__name__)
        log.setLevel(0)
        subpath = info["match"]["subpath"]
        log.debug("subpath is %r", subpath)
        if not subpath:
            log.debug("no subpath, returning false")
            return False
        parts = [self.subdir]
        parts.extend(subpath)
        resource_name = "/".join(parts)
        log.debug("package=%r, resource_name=%r", self.package, resource_name)
        return pkg_resources.resource_exists(self.package, resource_name)



import os
import shutil
import tempfile
import unittest

import sqlalchemy as sa
from sqlalchemy.engine.base import Engine

import pyramid_sqla as psa

class DBInfo(object):
    def __init__(self, dir, filename):
        self.file = os.path.join(dir, filename)
        self.url = "sqlite:///" + self.file

class PyramidSQLATestCase(unittest.TestCase):
    def setUp(self):
        self.dir = tempfile.mkdtemp()
        self.db1 = DBInfo(self.dir, "db1.sqlite")
        self.db2 = DBInfo(self.dir, "db2.sqlite")
        self.db3 = DBInfo(self.dir, "db3.sqlite")

    def tearDown(self):
        psa.reset()
        shutil.rmtree(self.dir, True)

    def assertDatabaseExists(self, db_file):
        self.assertEqual(os.path.exists(db_file), True)

    def assertDatabaseDoesNotExist(self, db_file):
        self.assertEqual(os.path.exists(db_file), False)

    if not hasattr(unittest.TestCase, "assertIsInstance"):
        def assertIsInstance(self, obj, classes):
            if not isinstance(obj, classes):
                typ = type(obj)
                if isinstance(classes, (list, tuple)):
                    classes_str = ", ".join(x.__name__ for x in classes)
                    classes_str = "[%s]" % classes_str
                else:
                    classes_str = classes.__name__
                msg = "%s is not an instance of %s" % (typ, classes_str)
                raise AssertionError(msg)

        def assertIs(self, a, b):
            if a is not b:
                raise AssertionError("%r is not %r" % (a, b))


class TestAddEngine(PyramidSQLATestCase):
    def test_keyword_args(self):
        engine = psa.add_engine(url=self.db1.url)
        self.assertIsInstance(engine, Engine)

    def test_simplest_settings(self):
        settings = {"sqlalchemy.url": self.db1.url}
        engine = psa.add_engine(settings, prefix="sqlalchemy.")
        self.assertIsInstance(engine, Engine)

    def test_existing_engine(self):
        e = sa.create_engine(self.db1.url)
        engine = psa.add_engine(engine=e)
        self.assertIs(engine, e)

    def test_multiple_engines(self):
        settings = {
            "sqlalchemy.url": self.db1.url,
            "stats.url": self.db2.url,
            "foo": "bar"}
        default = psa.add_engine(settings)
        stats = psa.add_engine(settings, name="stats", prefix="stats.")
        # Can we retrieve the engines?
        self.assertIs(psa.get_engine(), default)
        self.assertIs(psa.get_engine("default"), default)
        self.assertIs(psa.get_engine("stats"), stats)
        # Are the session binding and base binding set correctly?
        self.assertIs(psa.get_session().bind, default)
        self.assertIs(psa.get_base().metadata.bind, default)

    def test_multiple_engines_without_default(self):
        settings = {
            "db1.url": self.db1.url,
            "db2.url": self.db2.url,
            "foo": "bar"}
        db1 = psa.add_engine(settings, name="db1", prefix="db1.")
        db2 = psa.add_engine(settings, name="db2", prefix="db2.")
        # Can we retrieve the engines?
        self.assertIs(psa.get_engine("db1"), db1)
        self.assertIs(psa.get_engine("db2"), db2)
        # There should be no default engine
        self.assertIs(psa.get_session().bind, None)
        self.assertIs(psa.get_base().metadata.bind, None)
        self.assertRaises(RuntimeError, psa.get_engine)

class TestDeclarativeBase(PyramidSQLATestCase):
    def test1(self):
        import transaction
        Base = psa.get_base()
        class Person(Base):
            __tablename__ = "people"
            id = sa.Column(sa.Integer, primary_key=True)
            first_name = sa.Column(sa.Unicode(100), nullable=False)
            last_name = sa.Column(sa.Unicode(100), nullable=False)
        psa.add_engine(url=self.db1.url)
        Base.metadata.create_all()
        fred = Person(id=1, first_name=u"Fred", last_name=u"Flintstone")
        wilma = Person(id=2, first_name=u"Wilma", last_name=u"Flintstone")
        barney = Person(id=3, first_name=u"Barney", last_name=u"Rubble")
        betty = Person(id=4, first_name=u"Betty", last_name=u"Rubble")
        Session = psa.get_session()
        sess = Session()
        sess.add_all([fred, wilma, barney, betty])
        transaction.commit()
        sess.expunge_all()
        del fred, wilma, barney, betty
        # Can we get back a record?
        barney2 = sess.query(Person).get(3)
        self.assertEqual(barney2.id, 3)
        self.assertEqual(barney2.first_name, u"Barney")
        self.assertEqual(barney2.last_name, u"Rubble")
        sql = sa.select([Person.first_name])
        # Can we iterate the first names in reverse alphabetical order?
        q = sess.query(Person.first_name).order_by(Person.first_name.desc())
        result = [x.first_name for x in q]
        control = [u"Wilma", u"Fred", u"Betty", u"Barney"]
        self.assertEqual(result, control)

    def test1_without_transaction_manager(self):
        Base = psa.get_base()
        class Person(Base):
            __tablename__ = "people"
            id = sa.Column(sa.Integer, primary_key=True)
            first_name = sa.Column(sa.Unicode(100), nullable=False)
            last_name = sa.Column(sa.Unicode(100), nullable=False)
        psa.add_engine(url=self.db1.url)
        Base.metadata.create_all()
        fred = Person(id=1, first_name=u"Fred", last_name=u"Flintstone")
        wilma = Person(id=2, first_name=u"Wilma", last_name=u"Flintstone")
        barney = Person(id=3, first_name=u"Barney", last_name=u"Rubble")
        betty = Person(id=4, first_name=u"Betty", last_name=u"Rubble")
        Session = psa.get_session()
        Session.configure(extension=None)  # XXX Kludge for SQLAlchemy/ZopeTransactionExtension bug
        sess = Session()
        sess.add_all([fred, wilma, barney, betty])
        sess.commit()
        sess.expunge_all()
        del fred, wilma, barney, betty
        # Can we get back a record?
        barney2 = sess.query(Person).get(3)
        self.assertEqual(barney2.id, 3)
        self.assertEqual(barney2.first_name, u"Barney")
        self.assertEqual(barney2.last_name, u"Rubble")
        sql = sa.select([Person.first_name])
        # Can we iterate the first names in reverse alphabetical order?
        q = sess.query(Person.first_name).order_by(Person.first_name.desc())
        result = [x.first_name for x in q]
        control = [u"Wilma", u"Fred", u"Betty", u"Barney"]
        self.assertEqual(result, control)





import logging

import pyramid_sqla as psa
import sqlalchemy as sa
import sqlalchemy.orm as orm
import transaction

log = logging.getLogger(__name__)

Base = psa.get_base()
Session = psa.get_session()


#class MyModel(Base):
#    __tablename__ = 'models'
#
#    id = sa.Column(sa.Integer, primary_key=True)
#    name = sa.Column(sa.Unicode(255), nullable=False)



"""Helper functions

This module is available as the ``h`` variable in templates, and can also be
imported into handler modules or elsewhere. Put any common functions you want
to access in all templates here.

The WebHelpers package (http://python.org/pypi/WebHelpers) contains some
commonly used helpers for HTML tag creation, text formatting, number
formatting, etc.

The template globals (``h`` et al) are set in
``{{package}}.subscribers.add_renderer_globals()``.
"""

#from webhelpers.html import *
#from webhelpers.html.tags import *
